/*
 * COMPONENT_NAME: ldap.clients
 *
 * ABSTRACT: generic program to modify or add entries using LDAP
 *
 * ORIGINS: 202,27
 *
 * (C) COPYRIGHT International Business Machines Corp. 1997, 2003
 * All Rights Reserved
 * Licensed Materials - Property of IBM
 *
 * US Government Users Restricted Rights - Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 */
/*
 * Copyright (c) 1995 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

/* ldapchangepwd.c - generic program to modify a user's password */

#if defined(__OS400__)
#pragma comment (copyright, \
   "5722SS1  (C) Copyright IBM Corp. 1997, 2003. \
   All rights reserved. \
   US Government Users Restricted Rights - \
   Use, duplication or disclosure restricted \
   by GSA ADP Schedule Contract with IBM Corp. \
   Licensed Materials-Property of IBM")
#endif /* OS400 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#if !defined( WIN32 ) && !defined( __OS400__ )
#include <sys/file.h>
#endif
#include <fcntl.h>
#if !defined(WIN32)
#include <unistd.h>
#endif
#include <ldap.h>

#ifdef LDAP_SSL_MAX
#if defined(__OS400__)
#include <skit40/ibmgsk.h>
#else
#include <ldapssl.h>
#endif
#endif

#include <ldif.h>

#ifdef __OS400__
#include "QgldNLS.H"
#include "lber-int.h"
#endif

#define OLD_PASSWORD 1
#define NEW_PASSWORD  2

#if defined( WIN32 ) || defined( _WIN32 ) || defined( __OS400__ )
#define strcasecmp stricmp
#endif

#define safe_realloc( ptr, size ) ( ptr == NULL ? malloc( size ) : \
      realloc( ptr, size ))

#ifdef __OS400__
/* LIBASCII support...this must be the last header file!
   If using LIBASCII support convert string data to code page 850 */
#ifdef USING_LIBASCII
#include "ascii_a.h"
#pragma convert(850)
#endif
#endif

#define LDAPMOD_MAXLINE  4096

#ifdef NEEDPROTOS
static void freepmods( LDAPMod **pmods );
#endif /* NEEDPROTOS */

#ifndef __OS400__
/* Required for password prompting */
#ifdef _WIN32
#include <conio.h>
#else
/* termios.h is defined by POSIX */
#include <termios.h>
#endif

/* Arbitrary limit to supplied password length */
#define MAX_SUPPLIED_PW_LENGTH 255
char* getPassword( int type );
#endif

#ifndef _WIN32
#ifdef _GCC3
#include <errno.h>
#else
extern int errno;
#endif
#endif

/* global variables */
static char *prog;
static char *binddn = NULL;
static char *passwd = (char *) "";
static char *newpasswd = (char *) "";
static char *ldaphost = NULL;
static char *mech = NULL;
static int hoplimit = 10;
static int ldapport = LDAP_PORT;
static int verbose = 0;
static int contoper = 0;
static int referrals = LDAP_OPT_ON;
static int deref = LDAP_DEREF_NEVER;
static LDAP *ld;
static int ldapversion = LDAP_VERSION3;
static int DebugLevel = 0;
static FILE *fp = NULL;
static char *charset = NULL;
static int keys = 0;
static int tls = 0;
static int ssl = 0;
static char *keyfile = NULL;
static char *keyfile_pw = NULL;
static char *keyfile_dn = NULL;
static char *debug_ptr = NULL;

LDAPControl** serverControls = NULL;

/* Digest-MD5 */
static char *digest_username = NULL;
static char *digest_realm = NULL;

/*
** Display usage statement and exit.
*/
void usage() {
   fprintf(stderr, "\nSends modify password requests to an LDAP server.\n");
   fprintf(stderr, "usage:\n");
   fprintf(stderr, "    %s -D binddn -w currentpassword -n newpassword [options]\n", prog);
   fprintf(stderr, "      binddn: distinguished name of entry to modify\n");
   fprintf(stderr, "options:\n" );
   fprintf(stderr, "    -C charset      character set name to use, as registered with IANA\n");
   fprintf(stderr, "    -d level        set debug level in LDAP library\n");
   fprintf(stderr, "    -D binddn       bind dn\n");
   fprintf(stderr, "    -G realm        realm used for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -h host         LDAP server host name\n");
   fprintf(stderr, "    -K keyfile      file to use for keys\n");
   fprintf(stderr, "    -m mechanism    perform SASL bind with the given mechanism\n");
   fprintf(stderr, "    -M              Manage referral objects as normal entries.\n");
#ifdef __OS400__

   fprintf(stderr, "    -n newpassword  new password\n");
#else

   fprintf(stderr, "    -n newpassword  new password or '?' for non-echoed prompt.\n");
#endif

   fprintf(stderr, "    -N key_name     private key name to use in keyfile\n");
   fprintf(stderr, "    -O maxhops      maximum number of referrals to follow in a sequence\n");
   fprintf(stderr, "    -p port         LDAP server port number\n");
   fprintf(stderr, "    -P key_pw       keyfile password\n");
   fprintf(stderr, "    -R              do not chase referrals\n");
   fprintf(stderr, "    -U username     user name for DIGEST-MD5 bind mechanism\n");
   fprintf(stderr, "    -v              verbose mode\n");
   fprintf(stderr, "    -V version      LDAP protocol version (2 or 3; default is 3)\n");
#ifdef __OS400__

   fprintf(stderr, "    -w password     current bind password\n");
#else

   fprintf(stderr, "    -w password     current bind password or '?' for non-echoed prompt\n");
#endif
   fprintf(stderr, "    -y              set proxied id for proxied authorization operation \n");
   fprintf(stderr, "    -Y              use a secure ldap connection (TLS)\n");
   fprintf(stderr, "    -Z              use a secure ldap connection (SSL)\n");
   exit(1);
}

/*
**      Print a memory error message and exit.
*/
void mem_error(const char *varName) {
   fprintf(stderr,"%s: Memory allocation error occured.\n", varName);
   exit(1);
}

/*
** Parse command line arguments.
*/
void parse_arguments(int argc, char **argv) {
   int i, port = 0, rc = LDAP_SUCCESS;
   const char *optpattern = "RMYZv?h:V:p:D:w:d:n:m:O:K:P:N:C:G:U:y:";
   extern char *optarg;
   extern int optind;
   int manageDsa = LDAP_FALSE;
   char *proxyDN=NULL;
   /* proxied authorization control request */
   int Proxy_control_requested = LDAP_FALSE;
   int stringLen=0;


   fp = stdin;
   while ((i = getopt(argc, argv, optpattern)) != EOF) {
      switch ( i ) {
      case 'V':
         ldapversion = atoi(optarg);
         if ((ldapversion != LDAP_VERSION2) && (ldapversion != LDAP_VERSION3)) {
            fprintf(stderr, "Unsupported version level supplied.\n");
            usage();
         }
         break ;
      case 'O':         /* set maximum referral hop count  */
         hoplimit = atoi( optarg );
         break;
      case 'm':         /* use SASL bind mechanism  */
         if (!(mech = strdup ( optarg )))
            mem_error("mech");
         break;
      case 'h':   /* ldap host */
         if (!(ldaphost = strdup( optarg )))
            mem_error("host");
         break;
      case 'D':   /* bind DN */
         if (!(binddn = strdup( optarg )))
            mem_error("bindDN");
         break;
      case 'w':   /* password */
#ifndef __OS400__

         if (optarg && optarg[0] == '?') {
            passwd = getPassword(OLD_PASSWORD);

         } else
#endif
            if (!(passwd = strdup( optarg )))
               mem_error("newpassword");
         break;
      case 'n':   /* new password */
#ifndef __OS400__

         if (optarg && optarg[0] == '?') {
            newpasswd = getPassword(NEW_PASSWORD);

         } else
#endif
            if (!(newpasswd = strdup( optarg )))
               mem_error("newpassword");
         break;
      case 'd':  /* debug */
         DebugLevel = atoi(optarg);
         break;
      case 'p':  /* port */
         ldapport = atoi( optarg );
         port = 1;
         break;
      case 'R':    /* don't automatically chase referrals */
         referrals = LDAP_OPT_OFF;
         break;
      case 'M':   /* manage referral objects as normal entries */
         manageDsa = LDAP_TRUE;
         if ((rc = ldap_add_control (LDAP_MANAGEDSAIT_OID, 0 , NULL, LDAP_OPT_ON, &serverControls)) != LDAP_SUCCESS){
            fprintf( stderr, "ldap_add_control: %d, %s\n",
                     rc, ldap_err2string( rc ) );
            exit(rc);
         }
         break;
      case 'y':   /* Proxied authorization control */
            rc=ldap_proxy_dn_prefix(&proxyDN,optarg);
            if (rc == LDAP_SUCCESS) {
                stringLen=strlen(proxyDN);
                rc = ldap_add_control (LDAP_PROXY_CONTROL_OID,stringLen,
                      proxyDN,1,&serverControls);
               free(proxyDN);
              }
          if (rc != LDAP_SUCCESS) {
              fprintf(stderr, "Incorrect Proxied DN specified. \n");
              usage();
               }
           else  Proxy_control_requested=LDAP_TRUE;
          break;
      case 'v':   /* verbose mode */
         verbose++;
         break;
      case 'K':  /* keyfile */
         if (!(keyfile = strdup( optarg )))
            mem_error("keyfile name");
         keys = 1;
         break;
      case 'P':  /* keyfile password */
         if (!(keyfile_pw = strdup( optarg )))
            mem_error("keyfile password");
         break;
      case 'N':  /* keyfile label*/
         if (!(keyfile_dn = strdup( optarg )))
            mem_error("keyfile label");
         break;
      case 'Y':  /* TLS */
         if (ssl) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         tls = 1;
         break;
      case 'Z':  /* SSL */
         if (tls) {
            fprintf(stderr, "Cannot use both SSL and TLS at the same time.\n");
            usage();
         }
         ssl = 1;
         break;
      case 'C':  /* charset */
         if (!(charset = strdup( optarg )))
            mem_error("charset");
         break;
      case 'U':  /* DIGEST-MD5 Username */
         if (!(digest_username = strdup( optarg )))
            mem_error("Username");
         break;
      case 'G':  /* DIGEST-MD5 Realm */
         if (!(digest_realm = strdup( optarg )))
            mem_error("realm");
         break;
      case '?':
      default:
         usage();
      }
   }

   if ((! (tls || ssl)) && keys) /* Specified a key file, but no security protocol */
      ssl = 1;                  /* Use SSL by default */

   if ( manageDsa && ( ldapversion == LDAP_VERSION2 ) ) {
      fprintf( stderr, "\n-M option requires version 3. -M ignored.\n\n");
      manageDsa = LDAP_FALSE;
   }

   if (argc - optind != 0)
      usage();

   if ((strlen(passwd) < 1) || (strlen(newpasswd) < 1))
      usage();

   if (binddn == NULL)
      usage();

   /* Use default SSL port if none specified */
   if (( port == 0 ) && ( ssl ))
      ldapport = LDAPS_PORT;

   if ( ! DebugLevel )
      if ( ( debug_ptr = getenv ( "LDAP_DEBUG" ) ) )
         DebugLevel = atoi ( debug_ptr );
}

#ifndef __OS400__
/*  Get a password from the user but don't display it. */
char* getPassword( int type ) {
   char supplied_password[ MAX_SUPPLIED_PW_LENGTH + 1 ]; /* Buffer for password */

#ifdef _WIN32

   char in = '\0';                  /* Input character */
   int len = 0;                    /* Length of password */
#else

   struct termios echo_control;
   struct termios save_control;

   int fd = 0;                 /* File descriptor */
   int attrSet = 0;            /* Checked later for reset */

   /* Get the file descriptor associated with stdin. */
   fd = fileno( stdin );

   if (tcgetattr( fd, &echo_control ) != -1) {
      save_control = echo_control;
      echo_control.c_lflag &= ~( ECHO | ECHONL );

      if (tcsetattr( fd, TCSANOW, &echo_control ) == -1) {
         fprintf(stderr, "Internal error setting terminal attribute.\n");
         exit( errno );
      }

      attrSet = 1;
   }
#endif

   /* Prompt for a password. */
   if (type == OLD_PASSWORD)
      fputs( "Enter Old password ==> ", stdout );
   else
      fputs( "Enter New password ==> ", stdout );
   fflush( stdout );

#ifdef _WIN32
   /* Windows 9x/NT will always read from the console, i.e.,
      piped or redirected input will be ignored. */
   while ( in != '\r' && len <= MAX_SUPPLIED_PW_LENGTH ) {
      in = _getch();

      if (in != '\r') {
         supplied_password[len] = in;
         len++;

      } else {
         supplied_password[len] = '\0';
      }
   }
#else
   /* Get the password from stdin. */
   fgets( supplied_password, MAX_SUPPLIED_PW_LENGTH, stdin );

   /* Remove the newline at the end. */
   supplied_password[strlen( supplied_password ) - 1] = '\0';

#endif

#ifndef _WIN32
   /* Reset the terminal. */
   if (attrSet && tcsetattr( fd, TCSANOW, &save_control ) == -1) {
      fprintf(stderr, "Unable to reset the display.\n");
   }
#endif

   fprintf( stdout, "\n" );

   return ( supplied_password == NULL ) ? supplied_password : strdup( supplied_password );
}
#endif

/*
** Rebind callback function.
*/
int rebindproc(LDAP *ld, char **dnp, char **pwp, int *methodp, int freeit) {
   if ( !freeit ) {
      if ( mech != NULL)
         *methodp = LDAP_AUTH_SASL;
      else
         *methodp = LDAP_AUTH_SIMPLE;
      if ( binddn != NULL ) {
         *dnp = strdup( binddn );
         *pwp = strdup ( passwd );
      } else {
         *dnp = NULL;
         *pwp = NULL;
      }
   } else {
      free ( *dnp );
      free ( *pwp );
   }
   return ( LDAP_SUCCESS );
}

/*
** Connect and bind to server.
*/
void connect_to_server() {
   int failureReasonCode, rc, authmethod;
   struct berval ber;
   struct berval *server_creds;
   int msgidp = 0;
   int errcode = 0;
   int error = 0;
   int rc2 = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;
   LDAPControl **pwdPolicy_controls = NULL;
   LDAPControl **bind_client_controls = NULL;

   /* call ldap_ssl_client_init if V3 and SSL */
   if ((tls || ssl) && (ldapversion == LDAP_VERSION3)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }

      if (verbose)
         printf( "ldap_ssl_client_init( %s, %s, 0, &failureReasonCode )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
#ifdef __OS400__
      if (keyfile) {
         char *xstr = keyfile;
         unsigned long xstrlen = strlen( xstr ) + 1;
         if ( xlate_utf8_to_local( &xstr, &xstrlen, 0 ) != LDAP_SUCCESS ) {
            fprintf( stderr, "keyfile - xlate_utf8_to_local conversion failed\n");
            exit( 1 );
         }
         keyfile = xstr;
      }
      if (keyfile_pw) {
         char *xstr = keyfile_pw;
         unsigned long xstrlen = strlen( xstr ) + 1;
         if ( xlate_utf8_to_local( &xstr, &xstrlen, 0 ) != LDAP_SUCCESS ) {
            fprintf( stderr, "keyfile_pw - xlate_utf8_to_local conversion failed\n");
            exit( 1 );
         }
         keyfile_pw = xstr;
      }
#endif  /* __OS400__ */

      rc = ldap_ssl_client_init( keyfile, keyfile_pw, 0, &failureReasonCode ) ;
      if (rc != LDAP_SUCCESS) {
         fprintf( stderr,
                  "ldap_ssl_client_init failed! rc == %d, failureReasonCode == %d\n",
                  rc, failureReasonCode );
         exit( 1 ) ;
      }
   }

#ifdef __OS400__
   char *ldaphost_lcp = ldaphost;
   char *keyfile_dn_lcp = keyfile;
   unsigned long len;
   if (ldaphost) {
      len = (unsigned long)(strlen(ldaphost) + 1);
      if (xlate_utf8_to_local(&ldaphost_lcp, &len, 0) != LDAP_SUCCESS) {
         fprintf( stderr, "ldaphost - xlate_utf8_to_local conversion failed\n");
         exit( 1 );
      }
   }

   if (keyfile_dn) {
      len = (unsigned long)(strlen(keyfile_dn) + 1);
      if (xlate_utf8_to_local(&keyfile_dn_lcp, &len, 0) != LDAP_SUCCESS) {
         fprintf( stderr, "keyfile_dn - xlate_utf8_to_local conversion failed\n");
         exit( 1 );
      }
   } else {
      keyfile_dn_lcp = NULL;
   }
#endif  /* __OS400__ */

   /* open connection to server */
   if (ldapversion == LDAP_VERSION3) {
      if (ssl) {
         if (verbose)
            printf("ldap_ssl_init( %s, %d, %s )\n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport,
                   ((keyfile_dn) ? keyfile_dn : "NULL"));
#ifdef __OS400__

         ld = ldap_ssl_init( ldaphost_lcp, ldapport, keyfile_dn_lcp ) ;
#else

         ld = ldap_ssl_init( ldaphost, ldapport, keyfile_dn ) ;
#endif

         if (ld == NULL) {
            fprintf( stderr, "ldap_ssl_init failed\n" ) ;
            perror( "ldap_ssl_init" ) ;
            exit( 1 ) ;
         }
      } else {
         if (verbose)
            printf("ldap_init(%s, %d) \n",
                   ((ldaphost) ? ldaphost : "NULL"), ldapport);
#ifdef __OS400__

         if ((ld = ldap_init(ldaphost_lcp, ldapport)) == NULL) {
#else
         if ((ld = ldap_init(ldaphost, ldapport)) == NULL) {
#endif
            perror("ldap_ssl_init");
            exit(1);
         }
      }
   }
   else {
      if ( verbose )
         printf("ldap_open( %s, %d )\n",
                ((ldaphost) ? ldaphost : "NULL"), ldapport );

#ifdef __OS400__

      if (( ld = ldap_open( ldaphost_lcp, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
#else
      if (( ld = ldap_open( ldaphost, ldapport )) == NULL ) {
         perror( "ldap_open" );
         exit( 1 );
      }
#endif

   }

   /* set options */
   ldap_set_option (ld, LDAP_OPT_PROTOCOL_VERSION, (void * )&ldapversion);
   if (ldapversion == LDAP_VERSION3) {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )&DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )&deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)&hoplimit);
      ldap_set_option( ld, LDAP_OPT_REBIND_FN, (void *)rebindproc);
   } else {
      ldap_set_option (ld, LDAP_OPT_DEBUG, (void * )DebugLevel);
      ldap_set_option (ld, LDAP_OPT_DEREF, (void * )deref);
      ldap_set_option( ld, LDAP_OPT_REFHOPLIMIT, (void *)hoplimit);
   }
   ldap_set_option (ld, LDAP_OPT_REFERRALS, (void * )referrals);
   if (binddn != NULL)
      ldap_set_rebind_proc( ld, (LDAPRebindProc)rebindproc );
   if (charset != NULL) {
      if (ldap_set_iconv_local_charset(charset) != LDAP_SUCCESS) {
         fprintf(stderr, "unsupported charset %s\n", charset);
         exit(0);
      }
      ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_ON);
   }
#ifdef __OS400__
   ldap_set_option(ld, LDAP_OPT_UTF8_IO, (void *)LDAP_UTF8_XLATE_OFF);
#endif

   /* call ldap_ssl_start if V2 and SSL */
   if ( ssl && (ldapversion == LDAP_VERSION2)) {
      if ( keyfile == NULL ) {
         keyfile = getenv("SSL_KEYRING");
         if (keyfile != NULL) {
            keyfile = strdup(keyfile);
         }
      }
      if (verbose)
         printf( "ldap_ssl_start( ld, %s, %s, %s )\n",
                 ((keyfile) ? keyfile : "NULL"),
                 ((keyfile_pw) ? keyfile_pw : "NULL"),
                 ((keyfile_dn) ? keyfile_dn : "NULL"));
#ifdef LDAP_SSL_MAX

      rc = ibm_set_unrestricted_cipher_support();
      if (rc != 0) {
         fprintf( stderr, "Warning: ibm_gsk_set_unrestricted_cipher_support failed! rc == %d\n", rc );
      }
#endif
      rc = ldap_ssl_start(ld, keyfile, keyfile_pw, keyfile_dn);
      if ( rc != 0) {
         fprintf( stderr, "ldap_ssl_start failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_ssl_start: %s \n", errormsg);
         exit( 1 );
      }
   }

   if (tls) {                   /* Start the TLS session */
      if ((rc = ldap_start_tls_s_np(ld, NULL)) != LDAP_SUCCESS) {
         fprintf(stderr, "ldap_start_tls_s() failed! RC = %d\n", rc);
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_start_tls_s_np: %s\n", errormsg);
         exit(rc);
      }
   }

   /* bind to server */
   if (ldapversion == LDAP_VERSION3) {

      /*Set password policy control*/
      if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &pwdPolicy_controls)) != LDAP_SUCCESS){
         fprintf( stderr, "ldap_add_control: %d, %s\n",
                  rc, ldap_err2string( rc ) );
         exit(rc);
      }

      if ( ! mech ) /* Use simple bind */
      {
         ber.bv_len = strlen ( passwd );
         ber.bv_val = passwd;
         rc = ldap_sasl_bind(ld, binddn, LDAP_SASL_SIMPLE, &ber, pwdPolicy_controls, NULL, &msgidp);

         rc2 = rc;

         if (rc == LDAP_SUCCESS) {
            error = ldap_result( ld, msgidp, 1, (struct timeval *) NULL, &res );
            rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                                    NULL, &returnedControls, 1 );
         }
      }
      else /* Presence of mechanism means SASL bind */ {
         /*
         ** Special case for mech="EXTERNAL".  Unconditionally set bind DN
         ** and credentials to NULL.  This option should be used in tandem
         ** with SSL and client authentication.  For other SASL mechanisms,
         ** use the specified bind DN and credentials.
         */
         if (strcmp(mech, LDAP_MECHANISM_EXTERNAL) == 0) {
            rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
            if (rc != LDAP_SUCCESS ) {
               errormsg = ldap_err2string(ldap_get_errno(ld));
               fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
               fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
               exit ( rc );
            }
         } else {
            if (strcmp(mech, LDAP_MECHANISM_GSSAPI) == 0) {
               rc = ldap_sasl_bind_s (ld, NULL, mech, NULL, NULL, NULL, &server_creds);
               if (rc != LDAP_SUCCESS ) {
                  errormsg = ldap_err2string(ldap_get_errno(ld));
                  fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
                  fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
                  exit ( rc );
               }
            } else /* other SASL mechanisms */ {
               if (strcmp(mech,LDAP_MECHANISM_DIGEST_MD5) == 0) {
                  if (digest_username) {
                     if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_USER_NAME_OID, strlen(digest_username) , digest_username, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                        ldap_perror (ld, "ldap_add_control");
                        exit ( rc );
                     }
                  }
                  if (digest_realm) {
                     if ((rc = ldap_add_control (IBM_CLIENT_DIGEST_REALM_NAME_OID, strlen(digest_realm) , digest_realm, LDAP_TRUE, &bind_client_controls)) != LDAP_SUCCESS) {
                        ldap_perror (ld, "ldap_add_control");
                        exit ( rc );
                     }
                  }
               }

               ber.bv_len = strlen ( passwd );
               ber.bv_val = passwd;
               rc = ldap_sasl_bind_s (ld, binddn, mech, &ber, pwdPolicy_controls,
                                      bind_client_controls, &server_creds);

               if (rc != LDAP_SUCCESS ) {
                  errormsg = ldap_err2string(ldap_get_errno(ld));
                  fprintf(stderr,"ldap_sasl_bind_s: %s \n", errormsg);
                  fprintf(stderr,"ldap_sasl_bind_s: extended error code: %i\n", ldap_get_exterror(ld));
                  exit ( rc );
               }
            }
         }

         rc2 = rc;

         /* Get any controls sent by the server during the sasl_bind_s */
         if ( rc == LDAP_SUCCESS ) {
            rc = ldap_get_bind_controls( ld, &returnedControls );
         }
      }

      /*Free the password policy controls*/
      ldap_controls_free(pwdPolicy_controls);

      /*Free the client controls*/
      ldap_controls_free(bind_client_controls);

      if (rc == LDAP_SUCCESS) {
         if (returnedControls != NULL) {
            ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
         }
      }
      if (rc2 != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc2);
      } else if (rc != LDAP_SUCCESS) {
         errormsg = ldap_err2string(rc);
      } else if (errcode != LDAP_SUCCESS) {
         errormsg = ldap_err2string(errcode);
      }
      if (controlerr) {
         errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
      }
      if (controlwarn) {
         warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
      }
      if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
         fprintf(stderr,"ldap_simple_bind:  ");
         if (errormsg != NULL) {
            fprintf(stderr,errormsg);
            fprintf(stderr," --- ");
         }
         if (errormsgctrl != NULL) {
            fprintf(stderr,errormsgctrl);
            fprintf(stderr,"\n");
         }
         if (warnmsgctrl != NULL) {
            fprintf(stderr,warnmsgctrl, controlres);
            fprintf(stderr,"\n");
         }

      } else if (errormsg != NULL) {
         fprintf(stderr,"ldap_simple_bind:   %s \n", errormsg);
      }
      if (rc2 != LDAP_SUCCESS) {
         /*
         ** ldap_get_exterror only valuable for ssl communication.
         ** In this example, for LDAP v3, the bind is the first
         ** instance in which communication actually flows to the
         ** server.  So, if there is an ssl configuration error or
         ** other ssl problem, this will be the first instance where
         ** it will be detected.
         */
         if (ssl) {
            fprintf( stderr, "Attempted communication over SSL.\n");
            fprintf( stderr, "  The extended error is %d.\n", ldap_get_exterror(ld));
         }
         exit( rc2 );
      } else if (errcode != LDAP_SUCCESS) {
         exit( errcode );
      }

   } else /* LDAP V2 */ {
      authmethod = LDAP_AUTH_SIMPLE;

      rc = ldap_bind_s( ld, binddn, passwd, authmethod );

      if ( rc != LDAP_SUCCESS ) {
         errormsg = ldap_err2string(ldap_get_errno(ld));
         fprintf(stderr,"ldap_bind_s: %s \n", errormsg);
         exit( rc );
      }
   }
}


/*
** Add operation to the modify structure.
*/
void addmodifyop(LDAPMod ***pmodsp, int modop, const char *attr, char *value, int vlen) {
   LDAPMod * *pmods;
   int i, j;
   struct berval *bvp;

   /*
   ** Data can be treated as binary (wire ready) if one of the
   ** following applies:
   **     1) it was base64 encode
   **     2) charset is not defined  (wwd not __OS400__)
   **     3) read from an external file
   */

   i = 0;
   pmods = *pmodsp;
   if ( pmods != NULL ) {
      for ( ; pmods[ i ] != NULL; ++i ) {
         if ( strcasecmp( pmods[ i ]->mod_type, attr ) == 0 &&
              pmods[ i ]->mod_op == modop ) {
            break;
         }
      }
   }

   if ( pmods == NULL || pmods[ i ] == NULL ) {
      if (( pmods = (LDAPMod * *)safe_realloc( pmods, (i + 2) *
                                               sizeof( LDAPMod * ))) == NULL ) {
         perror( "safe_realloc" );
         exit( 1 );
      }
      *pmodsp = pmods;
      pmods[ i + 1 ] = NULL;
      if (( pmods[ i ] = (LDAPMod * )calloc( 1, sizeof( LDAPMod ))) == NULL ) {
         perror( "calloc" );
         exit( 1 );
      }
      pmods[ i ]->mod_op = modop;
      if (( pmods[ i ]->mod_type = strdup( attr )) == NULL ) {
         perror( "strdup" );
         exit( 1 );
      }
   }

   if ( value != NULL ) {
      if (modop & LDAP_MOD_BVALUES) {
         j = 0;
         if ( pmods[ i ]->mod_bvalues != NULL ) {
            for ( ; pmods[ i ]->mod_bvalues[ j ] != NULL; ++j ) {
               ;
            }
         }
         if (( pmods[ i ]->mod_bvalues =
               (struct berval **)safe_realloc( pmods[ i ]->mod_bvalues,
                                               (j + 2) * sizeof( struct berval *))) == NULL ) {
            perror( "safe_realloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_bvalues[ j + 1 ] = NULL;
         if (( bvp = (struct berval *)malloc( sizeof( struct berval )))
             == NULL ) {
            perror( "malloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_bvalues[ j ] = bvp;

         bvp->bv_len = vlen;
         if (( bvp->bv_val = (char *)malloc( vlen + 1 )) == NULL ) {
            perror( "malloc" );
            exit( 1 );
         }
         memmove( bvp->bv_val, value, vlen );
         bvp->bv_val[ vlen ] = '\0';
      } else {
         j = 0;
         if ( pmods[ i ]->mod_values != NULL ) {
            for ( ; pmods[ i ]->mod_values[ j ] != NULL; ++j ) {
               ;
            }
         }
         if (( pmods[ i ]->mod_values =
               (char **)safe_realloc( pmods[ i ]->mod_values,
                                      (j + 2) * sizeof( char *))) == NULL ) {
            perror( "safe_realloc" );
            exit( 1 );
         }
         pmods[ i ]->mod_values[ j + 1 ] = NULL;
         if (( pmods[ i ]->mod_values[ j ] = strdup( value )) == NULL) {
            perror( "strdup" );
            exit( 1 );
         }
      }
   }
}


/*
** Print a binary value. If charset is not specified then check to
** see if string is printable anyway.
*/
void print_binary(struct berval *bval) {
   int i, binary;

   printf( "\tBINARY (%ld bytes) ", bval->bv_len);
   if (charset == NULL) {
      binary = 0;
      for (i = 0; (i < bval->bv_len) && (!binary); ++i)
         if (!isprint(bval->bv_val[i]))
            binary = 1;
      if (!binary)
         for (i = 0; i < bval->bv_len; ++i)
            putchar(bval->bv_val[i]);
   }
   putchar('\n');
}


/*
** Modify or add an entry.
*/
int dopasswordmodify( char *dn, LDAPMod **pmods) {
   int i, j, op, rc, rc2;
   struct berval *bvp;
   int errcode = 0;
   int error = 0;
   int msgid = 0;
   LDAPMessage *res;
   int controlres = 0;
   int controlerr = 0;
   int controlwarn = 0;
   char *errmsg = NULL;
   const char *errormsg = NULL;
   const char *errormsgctrl = NULL;
   const char *warnmsgctrl = NULL;
   LDAPControl **returnedControls = NULL;


   if ( pmods == NULL ) {
      fprintf( stderr, "%s: no attributes to change or add (entry %s)\n",
               prog, dn );
      return ( LDAP_PARAM_ERROR );
   }

   printf( "changing password for entry %s\n", dn );

   if ( verbose ) {
      for ( i = 0; pmods[ i ] != NULL; ++i ) {
         op = pmods[ i ]->mod_op & ~LDAP_MOD_BVALUES;
         printf( "%s %s:\n", op == LDAP_MOD_REPLACE ?
                 "replace" : op == LDAP_MOD_ADD ?
                 "add" : "delete", pmods[ i ]->mod_type );
         if (pmods[i]->mod_op & LDAP_MOD_BVALUES) {
            if (pmods[ i ]->mod_bvalues != NULL) {
               for (j = 0; pmods[i]->mod_bvalues[j] != NULL; ++j)
                  print_binary(pmods[i]->mod_bvalues[j]);
            }
         } else {
            if (pmods[i]->mod_values != NULL) {
               for (j = 0; pmods[i]->mod_values[j] != NULL; ++j)
                  printf("\t%s\n", pmods[i]->mod_values[j]);
            }
         }
      }
   }

   if ((rc = ldap_add_control (LDAP_PWDPOLICY_CONTROL_OID, 0 , NULL, LDAP_OPT_OFF, &serverControls)) != LDAP_SUCCESS){
      fprintf( stderr, "ldap_add_control: %d, %s\n",
               rc, ldap_err2string( rc ) );
   }

   rc = ldap_modify_ext( ld, dn, pmods, serverControls,
                         NULL, &msgid);

   rc2 = rc;
   if (rc == LDAP_SUCCESS) {
      error = ldap_result( ld, msgid, 1, (struct timeval *) NULL, &res );
      rc = ldap_parse_result( ld, res, &errcode, NULL, &errmsg,
                              NULL, &returnedControls, 1 );
   }
   if (rc == LDAP_SUCCESS) {
      if (returnedControls != NULL) {
         ldap_parse_pwdpolicy_response(returnedControls, &controlerr, &controlwarn, &controlres);
      }
   }
   if (rc2 != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc2);
   } else if (rc != LDAP_SUCCESS) {
      errormsg = ldap_err2string(rc);
   } else if (errcode != LDAP_SUCCESS) {
      errormsg = ldap_err2string(errcode);
   }
   if (controlerr) {
      errormsgctrl = ldap_pwdpolicy_err2string(controlerr);
   }
   if (controlwarn) {
      warnmsgctrl = ldap_pwdpolicy_err2string(controlwarn);
   }
   if ((errormsgctrl != NULL) || (warnmsgctrl != NULL)) {
      if (errormsg != NULL) {
         fprintf(stderr,errormsg);
         fprintf(stderr," --- ");
      }
      if (errormsgctrl != NULL) {
         fprintf(stderr,errormsgctrl);
         fprintf(stderr,"\n");
      }
      if (warnmsgctrl != NULL) {
         fprintf(stderr,warnmsgctrl, controlres);
         fprintf(stderr,"\n");
      }
   } else if (errormsg != NULL) {
      fprintf(stderr," %s \n", errormsg);
   }
   if ( verbose ) {
      printf( "ldapchangepwd complete\n" );
   }
   putchar( '\n' );

   /* Return the proper error code */
   if (rc2)
	   return (rc2);
   else if (rc)
	   return (rc);
   else
	   return (errcode);
}


main( int argc, char **argv ) {
   char *p, buf[ 4096 ];
   int rc, i, authmethod, len, havedn;
   LDAPMod **pmods = NULL;
   FILE * cf_fd;
   const char *type = "userpassword";

   /* Strip off any path info on program name */
   if ((prog = strrchr(argv[0], '/')) != NULL)
      ++prog;
   else
      prog = argv[0];

#ifdef __OS400__

   prog = "ldapchangepwd";
#endif


   /* parse command line arguments */
   parse_arguments(argc, argv);


   /* connect to server */
   connect_to_server();

   rc = 0;
   addmodifyop(&pmods, LDAP_MOD_DELETE, type, passwd, strlen(passwd));
   addmodifyop(&pmods, LDAP_MOD_ADD, type, newpasswd, strlen(passwd));
   rc = dopasswordmodify(binddn, pmods);


   if ( tls )
      ldap_stop_tls_s_np ( ld );

   /* unbind and exit */
   ldap_unbind(ld);
   if(serverControls);
   ldap_controls_free(serverControls);
   exit(rc);
}

/*
 * NOTE: If you add new prototypes to this file, add an argopt pragma.
 */
#ifdef __OS400_ARG_OPT__
#pragma argopt(addmodifyop)
#pragma argopt(dopasswordmodify)
#endif /* __OS400_ARG_OPT__ */
